﻿// Goldie: GOLD Engine for D
// Samples: Static-Language Calculator
// Written in the D programming language.

/++
A simple arithmetic calculator utilizing a calculator grammar
compiled by GOLD.

Uses a statically loaded language and static-style tokens.

$(WEB http://www.dsource.org/projects/goldie, Goldie Homepage)

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

To re-generate the static language used by this app, run
the following command from the main Goldie directory:
    bin/staticlang lang/calc.cgt -dir:src -pack:samples.calculatorStatic.calc

This has been tested to work with:
  - DMD 1.056 / Tango 0.99.9 / Rebuild 0.76
  - DMD 1.056 / Tango 0.99.9 / xfBuild 0.4

+++++++++++++++++++++++++++++++++++++++++++++++
  
Note that, instead of the approach used in this program, we could also
have created one function that handles all rules:

----------------------
real calculate(Token_calc tok)
{
	if(auto t = cast(Token_calc!("Number"))tok)
		return to!(real)(t.toString());
		
	if(auto t = cast(Token_calc!("<Add Exp>", "<Mult Exp>"))tok)
		return calculate(t.sub!(0));
		
	// Etc...
	
	throw new Exception("Unhandled token: {}".sformat(tok.name));
}
----------------------

Or we could have used a templated function to increase the likelihood of
catching problems at compile-time:

----------------------
real calculate(T)(T tok)
{
	static if(!is( T : Token ))
		static assert(false, "T must be a Token_calc, not '"~T.stringof~"'");
		
	else static if(is( T == Token_calc!("<Add Exp>", "<Mult Exp>") ))
		return calculate(tok.sub!(0));
		
	else static if(is( T == Token_calc!("<Add Exp>", "<Add Exp>", "+", "<Mult Exp>") ))
		return calculate(tok.sub!(0)) + calculate(tok.sub!(2));
		
	else static if(is( T == Token_calc!("<Add Exp>", "<Add Exp>", "-", "<Mult Exp>") ))
		return calculate(tok.sub!(0)) - calculate(tok.sub!(2));
		
	// This is needed to determine what rule a given <Add Exp> was made from.
	else static if(is( T == Token_calc!("<Add Exp>") ))
	{
		auto t1 = cast(Token_calc!("<Add Exp>", "<Mult Exp>"))tok;
		auto t2 = cast(Token_calc!("<Add Exp>", "<Add Exp>", "+", "<Mult Exp>"))tok;
		auto t3 = cast(Token_calc!("<Add Exp>", "<Add Exp>", "-", "<Mult Exp>"))tok;
		
		if(t1)
			return calculate(t1);
		else if(t2)
			return calculate(t2);
		else if(t3)
			return calculate(t3);
		else
			throw new Exception("Unexpected form of <Add Exp>");
	}
		
	else
		static assert(false, "Unhandled type: "~T.StringOf);
}
----------------------

Or, of course, we could have done the above without templates:

----------------------
real calculate(Token_calc!("<Add Exp>", "<Mult Exp>") tok)
{
	return calculate(t.sub!(0)) + calculate(t.sub!(2));
}
real calculate(Token_calc!("<Add Exp>", "<Add Exp>", "+", "<Mult Exp>") tok)
{
	return calculate(t.sub!(0)) + calculate(t.sub!(2));
}
real calculate(Token_calc!("<Add Exp>", "<Add Exp>", "-", "<Mult Exp>") tok)
{
	return calculate(t.sub!(0)) - calculate(t.sub!(2));
}

// This is needed to determine what rule a given <Add Exp> was made from.
real calculate(Token_calc!("<Add Exp>") tok)
{
	// Same as in the templated example above
}
// Etc...
----------------------

+/

module samples.calculatorStatic.main;

import tango.io.Stdout;
import tango.text.Unicode;
import tango.util.Convert;

import semitwist.util.all : sformat;
import semitwist.cmd.plain;
import semitwist.util.compat.all;

import goldie.all;

import samples.calculatorStatic.calc.all;

real calculate(Token_calc!("Number") tok)
{
	return to!(real)(tok.toString());
}

real calculate(Token_calc!("<Add Exp>") tok)
{
	if(auto t = cast(Token_calc!("<Add Exp>", "<Mult Exp>"))tok)
		return calculate(t.sub!(0));
	
	if(auto t = cast(Token_calc!("<Add Exp>", "<Add Exp>", "+", "<Mult Exp>"))tok)
		return calculate(t.sub!(0)) + calculate(t.sub!(2));
	
	if(auto t = cast(Token_calc!("<Add Exp>", "<Add Exp>", "-", "<Mult Exp>"))tok)
		return calculate(t.sub!(0)) - calculate(t.sub!(2));
	
	throw new Exception("Unhandled token: {}".sformat(tok.name));
}

real calculate(Token_calc!("<Mult Exp>") tok)
{
	if(auto t = cast(Token_calc!("<Mult Exp>", "<Negate Exp>"))tok)
		return calculate(t.sub!(0));

	if(auto t = cast(Token_calc!("<Mult Exp>", "<Mult Exp>", "*", "<Negate Exp>"))tok)
		return calculate(t.sub!(0)) * calculate(t.sub!(2));

	if(auto t = cast(Token_calc!("<Mult Exp>", "<Mult Exp>", "/", "<Negate Exp>"))tok)
		return calculate(t.sub!(0)) / calculate(t.sub!(2));

	throw new Exception("Unhandled token: {}".sformat(tok.name));
}

real calculate(Token_calc!("<Negate Exp>") tok)
{
	if(auto t = cast(Token_calc!("<Negate Exp>", "<Value>"))tok)
		return calculate(t.sub!(0));

	if(auto t = cast(Token_calc!("<Negate Exp>", "-", "<Value>"))tok)
		return -calculate(t.sub!(1));

	throw new Exception("Unhandled token: {}".sformat(tok.name));
}

real calculate(Token_calc!("<Value>") tok)
{
	if(auto t = cast(Token_calc!("<Value>", "Number"))tok)
		return calculate(t.sub!(0));

	if(auto t = cast(Token_calc!("<Value>", "(", "<Add Exp>", ")"))tok)
		return calculate(t.sub!(1));

	throw new Exception("Unhandled token: {}".sformat(tok.name));
}

real calculate(Token_calc!() tok)
{
	throw new Exception("Unhandled token: {}".sformat(tok.name));
}

int main(string[] args)
{
	Stdout("Calculator").newline;
	auto lang = language_calc;
	// Note: At this point, the "dynamic" style could be used if desired, as
	//       demonstrated in the "calculatorDynamic" sample. But, since the
	//       language was loaded statically this time, we also have the option
	//       of using the "static" style, which is what this app demonstrates.

	Stdout("Enter an arithmetic expression ('exit' to exit)").newline;
	while(true)
	{
		Stdout.newline;
		string src = cmd.prompt("calculator>");
		
		if(src.toLower() == "exit")
			break;
			
		try
		{
			auto parseTree = lang.parseCode(src).parseTree;
			real result = calculate(parseTree);
			Stdout.formatln("{:d5}", result);
		}
		catch(Exception e)
			Stdout.formatln("{}", e);
	}
	
	return 0;
}
