﻿// Goldie: GOLD Engine for D
// Tools: Generate Documentation
// Written in the D programming language.

/++
Generate documentation.

$(WEB http://www.dsource.org/projects/goldie, Goldie Homepage)

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

This has been tested to work with:
  - DMD 1.056 / Tango 0.99.9 / Rebuild 0.76
  - DMD 1.056 / Tango 0.99.9 / xfBuild 0.4
+/

module tools.gendocs.main;

import semitwist.cmd.all;
import goldie.all;

// Requires an extra lib to be linked in on Linux, so just version it for Win for now.
version(Windows)
	debug import tango.core.tools.TraceExceptions;

import tools.gendocs.asttag;
import tools.gendocs.cmd;
import tools.gendocs.exception;
import tools.gendocs.page;

//TODO: See about allowing subtags in all types of statements.
//TODO: Create [$apiMetaParamHead|metaParamName|...$] (highlights metaparams).
//TODO: Highlight template param names (like in "staticSub!(int index)()")
//TODO: Fix: No file/line reported when [$file|...$] can't find the file.

int main(string[] args)
{
	Goldie.langDir = getExecPath()~"../lang/";

	auto cmdArgs = new CmdArgs(args);
	if(cmdArgs.shouldExit())
		return 1;

	Page.pageFilename = cmdArgs.outFile;
	Page.srcDir       = cmdArgs.srcDir;
	Page.astOutDir    = cmdArgs.astDir;
	Page.trimLinks    = cmdArgs.trimLinks;
	
	auto startDir = cmd.dir;
	
	if(!cmdArgs.quietMode)
		cmd.echo("Loading templates...");

	Page.vars = 
	[
		""[]: ""[]
	];

	string pageMap(Page p, Page mapRoot)
	{
		string str = "";
		foreach(Page subPage; mapRoot.sub)
		{
			str ~= "<li>";
			str ~= p.linkTo(subPage, "pagemenu-link");
			str ~= pageMap(p, subPage);
			str ~= "</li>";
		}
		if(str != "")
			str = "<ul>"~str~"</ul>";
		return str;
	}

	void ensureValidPage(string name, AST_Tag tag)
	{
		if(!(name in Page.pageLookup))
		{
			if(name in Page.ambiguousPageNames)
				throw new SemanticException(tag, "Ambiguous Page: '{}'".sformat(name));
			else
				throw new SemanticException(tag, "Unknown Page: '{}'".sformat(name));
		}
	}

	void ensureValidType(string name, AST_Tag tag)
	{
		if(!(name in Page.typeLookup))
			throw new SemanticException(tag, "Unknown Type: '{}'".sformat(name));
	}

	Page.dgVars = 
	[
		"title"[]   : (Page p, AST_Tag tag, string[] params) { return p.title;     },
		"name"      : (Page p, AST_Tag tag, string[] params) { return p.name;      },
		"desc"      : (Page p, AST_Tag tag, string[] params) { return p.desc;      },
		"fullName"  : (Page p, AST_Tag tag, string[] params) { return p.fullName;  },
		"urlToBase" : (Page p, AST_Tag tag, string[] params) { return p.urlToRoot; },
		
		"breadcrumbs" : (Page p, AST_Tag tag, string[] params)
		{
			string str = "";
			for(auto currPage = p; currPage !is null; currPage = currPage.parent)
			{
				if(str != "")
					str = " -> " ~ str;
					
				str = p.linkTo(currPage, "breadcrumb-link") ~ str;
			}

			return str;
		},
		
		"pageMenu" : (Page p, AST_Tag tag, string[] params)
		{
			string str = "";
			foreach(Page subPage; p.sub)
			{
				str ~= "<li>";
				str ~= p.linkTo(subPage, "pagemenu-link");
				if(subPage.desc !is null && subPage.desc != "")
				{
					str ~= " - ";
					str ~= subPage.desc;
				}
				str ~= "</li>";
			}
			str = "<ul>"~str~"</ul>";
			return str;
		},
		
		"pageMap" : (Page p, AST_Tag tag, string[] params)
		{
			return pageMap(p, p);
		},

		"siteMap" : (Page p, AST_Tag tag, string[] params)
		{
			Page rootPage = p;
			while(rootPage.parent !is null)
				rootPage = rootPage.parent;

			return
				p.linkTo(rootPage, "pagemenu-link") ~
				pageMap(p, rootPage);
		},

		"parentPageLink" : (Page p, AST_Tag tag, string[] params)
		{
			if(p.parent is null)
				return ""[];
			else
				return p.linkTo(p.parent);
		},

		"pageTitle" : (Page p, AST_Tag tag, string[] params)
		{
			ensureValidPage(params[0], tag);
			auto target = Page.pageLookup[params[0]];
			return p.linkTo(target, "page-link");
		},

		"pageName" : (Page p, AST_Tag tag, string[] params)
		{
			ensureValidPage(params[0], tag);
			auto target = Page.pageLookup[params[0]];
			return p.linkTo(target, "page-link", target.name);
		},

		"pageLink" : (Page p, AST_Tag tag, string[] params)
		{
			auto indexOfHash = params[0].locate('#');
			auto pageName = params[0][0..indexOfHash];
			auto anchor = (indexOfHash != params[0].length)? params[0][indexOfHash+1..$] : "";
			
			auto target = p;
			if(pageName != "")
			{
				ensureValidPage(pageName, tag);
				target = Page.pageLookup[pageName];
			}
			return p.linkTo(target, "page-link", params[1], anchor);
		},
		
		"typeName" : (Page p, AST_Tag tag, string[] params)
		{
			ensureValidType(params[0], tag);
			return p.linkToType(params[0], "page-link");
		},
		
		"typeLink" : (Page p, AST_Tag tag, string[] params)
		{
			ensureValidType(params[0], tag);
			string label = highlightMetaParams(params[1]);
			return p.linkToType(params[0], "page-link", label);
		},
		
		"apiHead" : (Page p, AST_Tag tag, string[] params)
		{
			return highlightAPIHead(p, tag, params);
		}
	];

	Page.lang = Goldie.loadLanguage("tmpl");
	
	Page page;
	try
		page = new Page(cmdArgs.srcFile);
	catch(Exception e)
	{
		if( cast(ParseException)e || cast(SemanticException)e )
		{
			cmd.echo(e.msg);
			
			if(cmdArgs.saveAST)
			{
				auto p = Page.partialObject;
				while(p.parent !is null)
					p = p.parent;
					
				if(!cmdArgs.quietMode)
					cmd.echo("Saving ASTs to JSON...");

				p.saveJSON();
			}
			
			return 1;
		}
		else
			throw e;
	}

	try
	{
		if(cmdArgs.saveAST)
		{
			if(!cmdArgs.quietMode)
				cmd.echo("Saving ASTs to JSON...");
				
			page.saveJSON();
		}

		cmd.dir = cmdArgs.outDir;
		Environment.cwd = cmd.dir.toString;

		if(!cmdArgs.quietMode)
			cmd.echo("Generating docs...");

		page.save();
	}
	catch(Exception e)
	{
		if( cast(ParseException)e || cast(SemanticException)e )
		{
			cmd.echo(e.msg);
			return 1;
		}
		else
			throw e;
	}
	
	if(!cmdArgs.quietMode)
		cmd.echo("Done!");

	return 0;
}

private string highlightAPIHead(Page p, AST_Tag tag, string[] params)
{
	string str = "";
	auto splitLines = params[0].trim().split("\n");
	foreach(string line; splitLines)
	{
		line = line.trim();
		if(line == "")
		{
			str ~= "<br />\n";
			continue;
		}
		
		// Split on left-parens that aren't preceeded by a '!'
		auto splitParens = regex.Regex(`(<[^!])\(`).split(line);
		
		// Separate member paramaters from the rest
		auto head = splitParens[0];
		splitParens = splitParens[1..$];
		
		// Extract member name
		int nameStart = head.locatePrior(' ') + 1; // UTF-safe because ' ' is one code-point
		if(nameStart > head.length)
			nameStart = 0;
		auto attrAndRet = head[0..nameStart].trim();
		auto name = head[nameStart..$];
		
		// Extract attribute list and return type
		int templStart = attrAndRet.locatePrior('!'); // UTF-safe because '!' is one code-point
		int retTypeStart = attrAndRet.locatePrior(' ', templStart) + 1; // UTF-safe because ' ' is one code-point
		if(retTypeStart > attrAndRet.length)
			retTypeStart = 0;
		auto attr = attrAndRet[0..retTypeStart].trim();
		auto retType = attrAndRet[retTypeStart..$];
		retType = highlightAPIHeadParams(p, tag, retType, false);
		
		// Output everything except params
		str ~=
			`<span class="attr">{}</span> {} <span class="ident">{}</span>`
				.sformat(attr, retType, name);
		
		// Output params
		if(splitParens.length > 0)
		{
			// Output template params
			foreach(paren; splitParens[0..$-1])
			{
				str ~= "(";
				str ~= highlightAPIHeadParams(p, tag, paren, false);
			}

			// Output regular params
			auto lastParen = splitParens[$-1];
			str ~= "(";
			str ~= highlightAPIHeadParams(p, tag, lastParen, true);
		}
		
		str ~= "<br />\n";
	}
	
	return `<div class="api-head">{}</div>`.sformat(str);
}

private string highlightAPIHeadParams(Page p, AST_Tag tag, string data, bool highlightVarNames)
{
	string str = "";
	auto paramSections = tokenizeParams(data);
	int parenDepth = 0;
	bool isAfterEquals = false;
	foreach(int paramSectionIndex, ParamTok section; paramSections)
	{
		switch(section.type)
		{
		case ParamTokType.Ident:
			if(highlightVarNames && parenDepth == 0)
			{
				if(paramSectionIndex < paramSections.length)
				{
					bool nextSectionFound = false;
					ParamTok nextSection;
					foreach(ParamTok potentialNextSection; paramSections[paramSectionIndex+1..$])
					if(potentialNextSection.type != ParamTokType.Whitespace)
					{
						nextSectionFound = true;
						nextSection = potentialNextSection;
						break;
					}
					
					auto nextType = nextSection.type;
					bool isParamName=false;
					if(nextType == ParamTokType.Equals)
						isParamName = true;
					else if(!isAfterEquals)
					{
						if( [ParamTokType.Comma, ParamTokType.CloseParen].contains(nextType) )
							isParamName = true;
						else if( nextType == ParamTokType.Misc )
						{
							if(nextSection.str.length >= 3 && nextSection.str[0..3] == "...")
								isParamName = true;
						}
					}
					
					if(isParamName)
					{
						str ~=
							`<span class="param-name">{}</span>`
								.sformat(section.str);
						continue;
					}
				}
			}
			
			if(section.str in Page.typeLookup)
				str ~= p.linkToType(section.str, "page-link");
			else
				str ~= section.str;
			break;
			
		case ParamTokType.Comma:
			isAfterEquals = false;
			str ~= section.str;
			break;
			
		case ParamTokType.Equals:
			isAfterEquals = true;
			str ~= section.str;
			break;
			
		case ParamTokType.OpenParen:
			parenDepth++;
			str ~= section.str;
			break;
			
		case ParamTokType.CloseParen:
			parenDepth--;
			str ~= section.str;
			break;
			
		case ParamTokType.Whitespace:
		case ParamTokType.Misc:
			str ~= section.str;
			break;

		default:
			throw new Exception("Internal Exception: Unhandled ParamTokType '{}'".sformat(section.type));
		}
	}
	
	return str;
}

enum ParamTokType
{
	Ident, Whitespace, Comma, OpenParen, CloseParen, Equals, Misc
}

struct ParamTok
{
	ParamTokType type;
	string str;
}

ParamTok[] tokenizeParams(string data)
{
	ParamTokType modeOf(dchar c)
	{
		switch(c)
		{
		case ',': return ParamTokType.Comma;
		case '(': return ParamTokType.OpenParen;
		case ')': return ParamTokType.CloseParen;
		case '=': return ParamTokType.Equals;
		
		default:
			if(isWhitespace(c))
				return ParamTokType.Whitespace;

			if(isLetterOrDigit(c) || "_{}"d.contains(c))
				return ParamTokType.Ident;
				
			return ParamTokType.Misc;
		}
	}
	
	ParamTok[] ret = [];
	dstring str = "";
	auto mode = ParamTokType.Whitespace;
	foreach(c; to!(dstring)(data))
	{
		auto nextMode = modeOf(c);
		if(mode != nextMode)
		{
			ret ~= ParamTok(mode, to!(string)(str));
			str = "";
		}
		str ~= c;
		mode = nextMode;
	}
	ret ~= ParamTok(mode, to!(string)(str));
	return ret;
}
